__all__ = []

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.padding import onnx_auto_pad_to_torch_padding

_CONV_CLASS_FROM_SPATIAL_RANK = {
    ('Conv', 1): nn.Conv1d,
    ('Conv', 2): nn.Conv2d,
    ('Conv', 3): nn.Conv3d,
    ('ConvTranspose', 1): nn.ConvTranspose1d,
    ('ConvTranspose', 2): nn.ConvTranspose2d,
    ('ConvTranspose', 3): nn.ConvTranspose3d,
}


@add_converter(operation_type='Conv', version=1)
@add_converter(operation_type='Conv', version=11)
@add_converter(operation_type='ConvTranspose', version=1)
@add_converter(operation_type='ConvTranspose', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    weights_value_name = node.input_values[1]
    weights = graph.initializers[weights_value_name]
    weights = weights.to_torch()
    if len(node.input_values) == 3:
        bias_value_name = node.input_values[2]
        bias = graph.initializers[bias_value_name]
        bias = bias.to_torch()
    else:
        bias = None

    op_type = node.operation_type
    spatial_rank = len(weights.shape) - 2
    try:
        conv_class = _CONV_CLASS_FROM_SPATIAL_RANK[op_type, spatial_rank]
    except KeyError as exc:
        raise NotImplementedError(
            f'Convolution operation with spatial rank == {spatial_rank} is not implemented'
        ) from exc

    node_attributes = node.attributes
    padding, input_padding_module = onnx_auto_pad_to_torch_padding(
        onnx_padding=node_attributes.get('pads', [0] * spatial_rank * 2),
        auto_pad=node_attributes.get('auto_pad', 'NOTSET'),
    )
    common_kwargs = {
        'kernel_size': node_attributes.get('kernel_shape', weights.shape[2:]),
        'stride': node_attributes.get('strides', 1),
        'dilation': node_attributes.get('dilations', 1),
        'groups': node_attributes.get('group', 1),
        'padding': padding,
        'bias': bias is not None,
    }

    if op_type == 'Conv':
        special_kwargs = {
            'out_channels': weights.shape[0],
            'in_channels': weights.shape[1] * common_kwargs['groups'],
        }
    elif op_type == 'ConvTranspose':
        if input_padding_module is not None:
            raise NotImplementedError('ConvTranspose with non symmetrical padding is not implemented.')

        output_padding = node_attributes.get('output_padding', [0] * spatial_rank)
        special_kwargs = {
            'out_channels': weights.shape[1] * common_kwargs['groups'],
            'in_channels': weights.shape[0],
            'output_padding': output_padding,
        }
    else:
        raise ValueError(f'Got unknown op_type "{op_type}"')

    torch_module = conv_class(
        **common_kwargs,
        **special_kwargs,
    )
    with torch.no_grad():
        torch_module.weight.data = weights
        if bias is not None:
            torch_module.bias.data = bias

    if input_padding_module is not None:
        torch_module = nn.Sequential(input_padding_module, torch_module)

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=OnnxMapping(
            inputs=(node.input_values[0],),
            outputs=node.output_values,
        ),
    )



import numpy as np
import torch
from onnx import numpy_helper
from onnx.onnx_ml_pb2 import TensorProto


class OnnxTensor:  # pylint: disable=missing-class-docstring
    def __init__(self, onnx_tensor_proto: TensorProto):
        self._proto = onnx_tensor_proto

    @classmethod
    def from_numpy(cls, array: np.ndarray, name: str = None):  # pylint: disable=missing-function-docstring
        onnx_tensor_proto = numpy_helper.from_array(array, name=name)
        return cls(onnx_tensor_proto)

    @classmethod
    def from_torch(cls, tensor: torch.Tensor, name: str = None):  # pylint: disable=missing-function-docstring
        array = tensor.detach().cpu().numpy()
        return cls.from_numpy(array, name=name)

    @property
    def proto(self) -> TensorProto:  # pylint: disable=missing-function-docstring
        return self._proto

    @property
    def name(self) -> str:  # pylint: disable=missing-function-docstring
        return self._proto.name

    def to_numpy(self) -> np.ndarray:  # pylint: disable=missing-function-docstring
        return numpy_helper.to_array(self._proto).copy()

    def to_torch(self) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return torch.from_numpy(self.to_numpy())



__all__ = [
    'OnnxMinMax',
]

import torch

from onnx2torch2.node_converters.base_element_wise import OnnxBaseElementWise
from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxMinMax(OnnxBaseElementWise):  # pylint: disable=missing-docstring
    _OPERATORS = {
        'Min': torch.amin,
        'Max': torch.amax,
    }

    def __init__(self, op_type: str):
        super().__init__(op_type=op_type)
        self._operator = self._OPERATORS[op_type]

    def apply_reduction(self, *tensors: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        broadcast_shape = self._broadcast_shape(*tensors)
        broadcast_tensors = [t.broadcast_to(broadcast_shape) for t in tensors]
        stacked_tensors = torch.stack(broadcast_tensors)
        output = self._operator(stacked_tensors, dim=0)
        return output


@add_converter(operation_type='Min', version=8)
@add_converter(operation_type='Min', version=12)
@add_converter(operation_type='Min', version=13)
@add_converter(operation_type='Max', version=8)
@add_converter(operation_type='Max', version=12)
@add_converter(operation_type='Max', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxMinMax(node.operation_type),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxPow',
    'OnnxSqrt',
]

from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import old_style_broadcast
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxPow(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, broadcast: Optional[int] = None, axis: Optional[int] = None):
        super().__init__()
        self.axis = axis
        self.broadcast = broadcast

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        exponent: torch.Tensor,
    ) -> torch.Tensor:
        if self.broadcast == 1 and self.axis is not None:
            exponent = old_style_broadcast(input_tensor, exponent, self.axis)

        return torch.pow(input_tensor, exponent)


class OnnxSqrt(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return torch.sqrt(input_tensor)


@add_converter(operation_type='Pow', version=1)
@add_converter(operation_type='Pow', version=7)
@add_converter(operation_type='Pow', version=12)
@add_converter(operation_type='Pow', version=13)
@add_converter(operation_type='Pow', version=15)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxPow(
            broadcast=node.attributes.get('broadcast', None),
            axis=node.attributes.get('axis', None),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='Sqrt', version=1)
@add_converter(operation_type='Sqrt', version=6)
@add_converter(operation_type='Sqrt', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxSqrt(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxScatterND',
]

from enum import Enum
from typing import Any
from typing import Dict

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class ReductionOnnxAttr(Enum):
    """
    Representation of new attribute in 16 opset: `reduction`.

    Type of reduction to apply: none (default), add, mul.
    - `none`: no reduction applied.
    - `add`: reduction using the addition operation.
    - `mul`: reduction using the multiplication operation.
    """

    NONE = 'none'
    ADD = 'add'
    MUL = 'mul'


class OnnxScatterND(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(self, reduction: ReductionOnnxAttr):
        super().__init__()

        if reduction != ReductionOnnxAttr.NONE:
            raise NotImplementedError(f'ScatterND with reduction attribute "{reduction.value}" is not implemented')

        self._reduction = reduction

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        onnx_attrs: Dict[str, Any] = {}

        if opset_version < 16:
            if self._reduction != ReductionOnnxAttr.NONE:
                raise ValueError(
                    'ScatterND from opset < 16 does not support'
                    f'reduction attribute != {ReductionOnnxAttr.NONE.value},'
                    f'got {self._reduction.value}'
                )
            return onnx_attrs

        onnx_attrs['reduction_s'] = self._reduction.value
        return onnx_attrs

    def forward(  # pylint: disable=missing-function-docstring
        self,
        data: torch.Tensor,
        indices: torch.Tensor,
        updates: torch.Tensor,
    ) -> torch.Tensor:
        def _forward():
            # There is no scatter nd for torch, use following formula:
            # https://github.com/onnx/onnx/blob/master/docs/Operators.md#ScatterND
            output = data.clone()  # to avoid RuntimeError: cannot mutate tensors with frozen storage
            ind_dim = indices.dim()
            # last dimension is a partial-index into data
            output_indices = torch.split(torch.t(indices.reshape((-1, indices.shape[-1]))), 1)
            # update.shape = indices.shape[0:ind_dim-1] ++ data.shape[indices.shape[-1]:data.dim()-1]
            output_updates = updates.reshape((-1, *updates.shape[ind_dim - 1 :]))
            output[output_indices] = output_updates

            return output.to(data.device)

        if torch.onnx.is_in_onnx_export():
            onnx_attrs = self._onnx_attrs(opset_version=get_onnx_version())
            return DefaultExportToOnnx.export(_forward, 'ScatterND', data, indices, updates, onnx_attrs)

        return _forward()


@add_converter(operation_type='ScatterND', version=11)
@add_converter(operation_type='ScatterND', version=13)
@add_converter(operation_type='ScatterND', version=16)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    reduction = ReductionOnnxAttr(node_attributes.get('reduction', 'none'))

    return OperationConverterResult(
        torch_module=OnnxScatterND(reduction=reduction),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



import tempfile
from pathlib import Path
from typing import Union

import onnx
from onnx.onnx_ml_pb2 import ModelProto
from onnx.shape_inference import infer_shapes
from onnx.shape_inference import infer_shapes_path


def _is_big_model(model: ModelProto) -> bool:
    return model.ByteSize() / (1024 * 1024 * 1024) > 2.0


def _shape_inference_by_model_path(
    model_path: Union[Path, str],
    output_path: Union[Path, str],
    **kwargs,
) -> ModelProto:
    model_path = str(Path(model_path).resolve())
    output_path = str(Path(output_path).resolve())
    infer_shapes_path(model_path, output_path=output_path, **kwargs)

    return onnx.load(output_path)


def safe_shape_inference(  # pylint: disable=missing-function-docstring
    onnx_model_or_path: Union[ModelProto, Path, str],
    **kwargs,
) -> ModelProto:
    if isinstance(onnx_model_or_path, ModelProto):
        if not _is_big_model(onnx_model_or_path):
            return infer_shapes(onnx_model_or_path, **kwargs)

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_model_path = Path(tmp_dir) / 'model.onnx'
            onnx.save_model(
                proto=onnx_model_or_path,
                f=str(tmp_model_path),
                save_as_external_data=True,
                all_tensors_to_one_file=True,
            )
            return _shape_inference_by_model_path(tmp_model_path, output_path=tmp_model_path, **kwargs)

    with tempfile.NamedTemporaryFile(dir=Path(onnx_model_or_path).parent) as tmp_model_file:
        return _shape_inference_by_model_path(onnx_model_or_path, output_path=tmp_model_file.name, **kwargs)



__all__ = [
    'OnnxConstant',
]

from typing import Any

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node

_CONSTANT_PARSING_MAPPING = {
    'value': lambda x: x.to_torch(),
    'value_float': torch.tensor,
    'value_floats': torch.tensor,
    'value_int': torch.tensor,
    'value_ints': torch.tensor,
    'value_string': lambda x: x,
    'value_strings': lambda x: x,
}


class OnnxConstant(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, value: Any):
        super().__init__()
        # We need it for placing constant to cuda.
        if isinstance(value, torch.Tensor):
            self.register_buffer('value', value)
        else:
            self.value = value

    def forward(self) -> Any:  # pylint: disable=missing-function-docstring
        return self.value


def _prepare_output_value(value: Any, attr_name: str) -> Any:
    if attr_name in _CONSTANT_PARSING_MAPPING:
        return _CONSTANT_PARSING_MAPPING[attr_name](value)

    raise NotImplementedError(f'value type "{attr_name}" not supported yet.')


@add_converter(operation_type='Constant', version=9)
@add_converter(operation_type='Constant', version=11)
@add_converter(operation_type='Constant', version=12)
@add_converter(operation_type='Constant', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    attr_name, value = list(node.attributes.items())[0]
    prepared_value = _prepare_output_value(value, attr_name)

    torch_module = OnnxConstant(
        value=prepared_value,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxWhere',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxWhere(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def forward(  # pylint: disable=missing-function-docstring
        self,
        condition: torch.Tensor,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        return torch.where(condition, x, y)


@add_converter(operation_type='Where', version=9)
@add_converter(operation_type='Where', version=16)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxWhere(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



# pylint: disable=missing-docstring
__all__ = [
    'OnnxArgExtremumOld',
    'OnnxArgExtremum',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node

DEFAULT_AXIS = 0
DEFAULT_KEEPDIMS = 1
DEFAULT_SELECT_LAST_INDEX = 0

_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'ArgMax': torch.argmax,
    'ArgMin': torch.argmin,
}


class OnnxArgExtremumOld(nn.Module, OnnxToTorchModule):
    def __init__(self, operation_type: str, axis: int, keepdims: int):
        super().__init__()
        self.axis = axis
        self.keepdims = bool(keepdims)
        self.extremum_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

    def forward(self, data: torch.Tensor) -> torch.Tensor:
        return self.extremum_function(data, dim=self.axis, keepdim=self.keepdims)


class OnnxArgExtremum(nn.Module, OnnxToTorchModule):
    def __init__(self, operation_type: str, axis: int, keepdims: int, select_last_index: int):
        super().__init__()
        self.axis = axis
        self.keepdims = bool(keepdims)
        self.select_last_index = bool(select_last_index)
        self.extremum_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

    def forward(self, data: torch.Tensor) -> torch.Tensor:
        if self.select_last_index:
            # torch's argmax does not handle the select_last_index attribute from Onnx.
            # We flip the data, call the normal argmax, then map it back to the original
            flipped = torch.flip(data, dims=[self.axis])

            extremum_index_flipped = self.extremum_function(flipped, dim=self.axis, keepdim=self.keepdims)
            extremum_index_original = data.size(dim=self.axis) - 1 - extremum_index_flipped
            return extremum_index_original

        return self.extremum_function(data, dim=self.axis, keepdim=self.keepdims)


@add_converter(operation_type='ArgMax', version=12)
@add_converter(operation_type='ArgMax', version=13)
@add_converter(operation_type='ArgMin', version=12)
@add_converter(operation_type='ArgMin', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    return OperationConverterResult(
        torch_module=OnnxArgExtremum(
            operation_type=node.operation_type,
            axis=node.attributes.get('axis', DEFAULT_AXIS),
            keepdims=node.attributes.get('keepdims', DEFAULT_KEEPDIMS),
            select_last_index=node.attributes.get('select_last_index', DEFAULT_SELECT_LAST_INDEX),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='ArgMax', version=11)
@add_converter(operation_type='ArgMin', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    return OperationConverterResult(
        torch_module=OnnxArgExtremumOld(
            operation_type=node.operation_type,
            axis=node.attributes.get('axis', DEFAULT_AXIS),
            keepdims=node.attributes.get('keepdims', DEFAULT_KEEPDIMS),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = []

from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


@add_converter(operation_type='LRN', version=13)
@add_converter(operation_type='LRN', version=1)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    size = node.attributes.get('size')
    alpha = node.attributes.get('alpha', 0.0001)
    beta = node.attributes.get('beta', 0.75)
    k = node.attributes.get('bias', 1)  # pylint: disable=invalid-name

    return OperationConverterResult(
        torch_module=nn.LocalResponseNorm(size=size, alpha=alpha, beta=beta, k=k),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'CustomExportToOnnx',
    'DefaultExportToOnnx',
    'OnnxToTorchModuleWithCustomExport',
]

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

import torch
from torch import _C as torch_C

from onnx2torch2.utils.common import OnnxToTorchModule


class OnnxToTorchModuleWithCustomExport(OnnxToTorchModule):
    """
    Marker class for onnx2torch modules with custom export to onnx.
    """

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:  # pylint: disable=unused-argument
        """
        Returns ONNX attributes with their values as a dictionary.

        Parameters
        ----------
        opset_version : int
            ONNX opset version.
            The number of attributes, their names and values depend on opset version;
            function should return correct set of attributes.

        Returns
        -------
        Dict[str, Any]
            ONNX attributes.

        """
        return {}


class CustomExportToOnnx(torch.autograd.Function):
    """Customizes ONNX exporting from PyTorch."""

    _NEXT_FORWARD_FUNCTION: Optional[Callable] = None

    @classmethod
    def export(cls, forward_function: Callable, *args) -> Any:
        """
        Substitues custom forward function.
        This function is closely related to forward function, it substitues `forward_function` to real forward.

        Old name: `set_forward_and_apply`.
        """
        CustomExportToOnnx._NEXT_FORWARD_FUNCTION = forward_function
        return cls.apply(*args)

    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Any:  # pylint: disable=unused-argument, arguments-differ
        """Applies custom forward function."""
        if CustomExportToOnnx._NEXT_FORWARD_FUNCTION is None:
            raise RuntimeError('Forward function is not set')

        try:
            return CustomExportToOnnx._NEXT_FORWARD_FUNCTION()  # pylint: disable=not-callable
        finally:
            CustomExportToOnnx._NEXT_FORWARD_FUNCTION = None

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:  # pylint: disable=unused-argument, missing-function-docstring
        raise RuntimeError('Backward called while converting to ONNX')

    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:  # pylint: disable=unused-argument
        """Export implementation. Return ONNX operation from this function using graph."""
        raise NotImplementedError


class DefaultExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    """
    CustomExportToOnnx with default symbolic method implementation.

    Please follow our convention, args consists of:
        - op_type
        - operation inputs
        - operation attributes

    DO NOT REORDER!

    Note: the number of operation outputs can be added later.

    This class should be used in most cases:
    >>> return DefaultExportToOnnx.export(_forward, op_type, *inputs, onnx_attrs)
    """

    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        op_type, *inputs, onnx_attrs = args
        return graph.op(op_type, *inputs, **onnx_attrs, outputs=1)



from typing import List
from typing import NamedTuple
from typing import Tuple
from typing import Union

import torch
from onnx import ValueInfoProto  # pylint: disable=no-name-in-module
from torch import nn
from torch.onnx import symbolic_helper

from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode


class OnnxToTorchModule:
    """
    Marker class for onnx2torch modules.
    """

    pass  # pylint: disable=unnecessary-pass


class OnnxMapping(NamedTuple):  # pylint: disable=missing-class-docstring
    inputs: Tuple[str, ...]
    outputs: Tuple[str, ...]


class OperationConverterResult(NamedTuple):  # pylint: disable=missing-class-docstring
    torch_module: nn.Module
    onnx_mapping: OnnxMapping


def onnx_mapping_from_node(node: OnnxNode) -> OnnxMapping:  # pylint: disable=missing-function-docstring
    return OnnxMapping(
        inputs=node.input_values,
        outputs=node.output_values,
    )


def get_onnx_version() -> int:
    """Returns opset version at the time of the export."""
    if hasattr(symbolic_helper, 'GLOBALS'):
        return symbolic_helper.GLOBALS.export_onnx_opset_version

    return symbolic_helper._export_onnx_opset_version  # pylint: disable=no-member, protected-access


def get_shape_from_value_info(value_info: ValueInfoProto) -> List[int]:  # pylint: disable=missing-function-docstring
    return [dim.dim_value for dim in value_info.type.tensor_type.shape.dim]


def get_const_value(  # pylint: disable=missing-function-docstring
    name: str,
    graph: OnnxGraph,
) -> Union[torch.Tensor, float, int, str, List]:
    if name in graph.initializers:
        return graph.initializers[name].to_torch()

    try:
        node, _ = graph.value_as_node_output(name)
    except KeyError as exc:
        raise KeyError(f'Tensor "{name}" is not found in constant values') from exc

    if node.operation_type == 'Constant':
        attr_name, attr_value = next(iter(node.attributes.items()))
        if attr_name == 'value':
            attr_value = attr_value.to_torch()

        return attr_value

    raise KeyError(f'Tensor "{name}" is not found in constant values')


def old_style_broadcast(  # pylint: disable=missing-function-docstring
    first: torch.Tensor,
    second: torch.Tensor,
    axis: int,
) -> torch.Tensor:
    rank = len(first.shape)
    axis = axis + rank if axis < 0 else axis

    second_shape = [1] * axis + list(second.shape)
    second_shape = second_shape + [1] * (rank - len(second_shape))

    return second.view(second_shape)



# pylint: disable=missing-docstring
import torch
from torch import nn

from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxBaseElementWise(nn.Module, OnnxToTorchModuleWithCustomExport):
    def __init__(self, op_type: str):
        super().__init__()
        self._op_type = op_type

    @staticmethod
    def _broadcast_shape(*tensors: torch.Tensor):
        shapes = [t.shape for t in tensors]
        broadcast_shape = torch.broadcast_shapes(*shapes)
        return broadcast_shape

    def apply_reduction(self, *tensors: torch.Tensor) -> torch.Tensor:
        del tensors
        raise NotImplementedError

    def forward(self, *input_tensors: torch.Tensor) -> torch.Tensor:
        if len(input_tensors) == 1:
            # If there is a single element, return it (no op).
            # Also, no need for manually building the ONNX node.
            return input_tensors[0]

        def _forward() -> torch.Tensor:
            return self.apply_reduction(*input_tensors)

        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(_forward, self._op_type, *input_tensors, {})

        return _forward()



from onnx2torch2.node_converters.activations import *
from onnx2torch2.node_converters.arg_extrema import *
from onnx2torch2.node_converters.average_pool import *
from onnx2torch2.node_converters.batch_norm import *
from onnx2torch2.node_converters.binary_math_operations import *
from onnx2torch2.node_converters.cast import *
from onnx2torch2.node_converters.clip import *
from onnx2torch2.node_converters.comparisons import *
from onnx2torch2.node_converters.concat import *
from onnx2torch2.node_converters.constant import *
from onnx2torch2.node_converters.constant_of_shape import *
from onnx2torch2.node_converters.conv import *
from onnx2torch2.node_converters.cumsum import *
from onnx2torch2.node_converters.depth_to_space import *
from onnx2torch2.node_converters.dropout import *
from onnx2torch2.node_converters.einsum import *
from onnx2torch2.node_converters.expand import *
from onnx2torch2.node_converters.eye_like import *
from onnx2torch2.node_converters.flatten import *
from onnx2torch2.node_converters.functions import *
from onnx2torch2.node_converters.gather import *
from onnx2torch2.node_converters.gemm import *
from onnx2torch2.node_converters.global_average_pool import *
from onnx2torch2.node_converters.gridsample import *
from onnx2torch2.node_converters.identity import *
from onnx2torch2.node_converters.instance_norm import *
from onnx2torch2.node_converters.layer_norm import *
from onnx2torch2.node_converters.logical import *
from onnx2torch2.node_converters.lrn import *
from onnx2torch2.node_converters.matmul import *
from onnx2torch2.node_converters.max_pool import *
from onnx2torch2.node_converters.mean import *
from onnx2torch2.node_converters.min_max import *
from onnx2torch2.node_converters.mod import *
from onnx2torch2.node_converters.neg import *
from onnx2torch2.node_converters.nms import *
from onnx2torch2.node_converters.pad import *
from onnx2torch2.node_converters.pow import *
from onnx2torch2.node_converters.range import *
from onnx2torch2.node_converters.reciprocal import *
from onnx2torch2.node_converters.reduce import *
from onnx2torch2.node_converters.registry import OperationDescription
from onnx2torch2.node_converters.registry import TConverter
from onnx2torch2.node_converters.registry import get_converter
from onnx2torch2.node_converters.reshape import *
from onnx2torch2.node_converters.resize import *
from onnx2torch2.node_converters.roialign import *
from onnx2torch2.node_converters.roundings import *
from onnx2torch2.node_converters.scatter_nd import *
from onnx2torch2.node_converters.shape import *
from onnx2torch2.node_converters.slice import *
from onnx2torch2.node_converters.split import *
from onnx2torch2.node_converters.squeeze import *
from onnx2torch2.node_converters.sum import *
from onnx2torch2.node_converters.tile import *
from onnx2torch2.node_converters.topk import *
from onnx2torch2.node_converters.transpose import *
from onnx2torch2.node_converters.unsqueeze import *
from onnx2torch2.node_converters.where import *



__all__ = [
    'OnnxTranspose',
]

from typing import List
from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult


class OnnxTranspose(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, perm: Optional[List[int]] = None):
        super().__init__()
        self.perm = perm

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self.perm is None:
            self.perm = list(range(input_tensor.dim()))[::-1]

        return input_tensor.permute(self.perm)


@add_converter(operation_type='Transpose', version=1)
@add_converter(operation_type='Transpose', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    input_values = [node.input_values[0]]
    perm_value_name = node.input_values[1] if len(node.input_values) > 1 else None

    if perm_value_name is not None:
        perm = graph.initializers[perm_value_name].to_torch().tolist()
    else:
        perm = node.attributes.get('perm', None)
        if perm is not None:
            perm = torch.tensor(perm, dtype=torch.long).tolist()

    return OperationConverterResult(
        torch_module=OnnxTranspose(perm=perm),
        onnx_mapping=OnnxMapping(
            inputs=tuple(input_values),
            outputs=node.output_values,
        ),
    )



__all__ = [
    'OnnxDropoutDynamic',
]

from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxDropoutDynamic(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def forward(  # pylint: disable=missing-function-docstring, unused-argument
        self,
        input_tensor: torch.Tensor,
        ratio: float = 0.5,
        training_mode: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # Ignoring training_mode from ONNX and use the one from PyTorch
        return F.dropout(input_tensor, p=ratio, training=self.training)


@add_converter(operation_type='Dropout', version=10)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    ratio = node_attributes.get('ratio', 0.5)

    torch_module = nn.Dropout(p=ratio)

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='Dropout', version=12)
@add_converter(operation_type='Dropout', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    seed = node_attributes.get('seed')
    if seed is not None:
        raise NotImplementedError('Dropout nodes with seeds are not supported.')

    return OperationConverterResult(
        torch_module=OnnxDropoutDynamic(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



# Copyright 2023 Untether AI Inc. ALL RIGHTS RESERVED.
#
# This source code (Information) is proprietary to Untether AI Inc.
# (Untether) and MAY NOT be copied by any method, incorporated into or
# linked to another program, reverse engineered or otherwise used in
# any manner whatsoever without the express prior written consent of
# Untether. This Information and any portion thereof is and shall
# remain the property of Untether. Untether makes no representation
# or warranty regarding the accuracy of the Information nor its fitness nor
# merchantability for any purpose.  Untether assumes no responsibility
# or liability for its use in any way and conveys to the user no license,
# right or title under any patent, copyright or otherwise, and makes
# no representation or warranty that this Information does not infringe
# any third party patent, copyright or other proprietary right.

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class UGridSampler(nn.Module):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        align_corners: int = 0,
        mode: int = 0,
        padding_mode: int = 0,
    ):
        super().__init__()
        self.align_corners = bool(align_corners)
        self.interpolation_mode = mode if mode != "linear" else "bilinear"
        self.padding_mode = padding_mode

    def forward(  # pylint: disable=missing-function-docstring
        self,
        x,
        grid,
    ) -> torch.Tensor:
        return torch.nn.functional.grid_sample(
            x, grid, mode=self.interpolation_mode, padding_mode=self.padding_mode, align_corners=self.align_corners
        )


@add_converter(operation_type='GridSample', version=16)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    align_corners = node_attributes.get('align_corners', 0)
    padding_mode = node_attributes.get('padding_mode', 0)
    mode = node_attributes.get('mode', 0)
    print(align_corners, padding_mode, mode)

    torch_module = UGridSampler(align_corners, mode, padding_mode)
    return OperationConverterResult(torch_module=torch_module, onnx_mapping=onnx_mapping_from_node(node))



__all__ = [
    'OnnxTile',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxTile(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(self, repeats: list):
        super().__init__()
        self.repeats = repeats

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        repeats: torch.Tensor,
    ) -> torch.Tensor:
        # torch.tile(input_tensor, repeats) is not supported for exporting
        repeats = self.repeats if self.repeats else repeats
        forward_lambda = lambda: input_tensor.repeat(torch.Size(repeats))
        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(forward_lambda, 'Tile', input_tensor, repeats, {})

        return forward_lambda()


@add_converter(operation_type='Tile', version=6)
@add_converter(operation_type='Tile', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    repeats = []
    if node.input_values[1] in graph.initializers:
        repeats = graph.initializers[node.input_values[1]].to_torch().cpu().tolist()
    return OperationConverterResult(
        torch_module=OnnxTile(repeats),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxGather',
    'OnnxGatherElements',
    'OnnxGatherND',
]

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport
from onnx2torch2.utils.indices import upcast_indices


class OnnxGatherElements(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, axis: int = 0):
        super().__init__()
        self._axis = axis

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        indices: torch.Tensor,
    ) -> torch.Tensor:
        return torch.gather(input_tensor, dim=self._axis, index=upcast_indices(indices))


class OnnxGather(nn.Module, OnnxToTorchModuleWithCustomExport):
    """ONNX Gather implementation (or numpy.take implementation)."""

    def __init__(self, axis: int = 0):
        super().__init__()
        self._axis = axis

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        return {'axis_i': self._axis}

    @staticmethod
    def slice_from_axis(  # pylint: disable=missing-docstring
        input_tensor: torch.Tensor,
        axis: int,
        indices: torch.Tensor,
    ) -> Tuple[Union[slice, torch.Tensor], ...]:
        if indices.dim() == 0:
            indices = indices.unsqueeze(0)
        axis = input_tensor.dim() + axis if axis < 0 else axis
        skip_axis: List[Union[slice, torch.Tensor]] = [slice(None)] * axis
        skip_axis.append(upcast_indices(indices))
        return tuple(skip_axis)

    def forward(  # pylint: disable=missing-function-docstring
        self, input_tensor: torch.Tensor, indices: torch.Tensor
    ) -> torch.Tensor:
        def _forward():
            # pytorch Gather differs from onnx Gather, onnx gather work like numpy.take
            # But torch.take does not support different axis. So we make it by yourself
            # numpy.take is input_data[:, :, indices] where we pass NONE slices AXIS time
            slice_for_take = self.slice_from_axis(input_tensor, self._axis, indices)
            if indices.dim() == 0:
                return input_tensor[slice_for_take].squeeze(self._axis)

            return input_tensor[slice_for_take]

        if torch.onnx.is_in_onnx_export():
            onnx_attrs = self._onnx_attrs(opset_version=get_onnx_version())
            return DefaultExportToOnnx.export(_forward, 'Gather', input_tensor, indices, onnx_attrs)

        return _forward()


class OnnxGatherND(nn.Module, OnnxToTorchModuleWithCustomExport):
    """ONNX GatherND implementation."""

    def __init__(self, batch_dims: int = 0):
        super().__init__()
        self._batch_dims: int = batch_dims

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        onnx_attrs: Dict[str, Any] = {}

        if opset_version == 11:
            if self._batch_dims != 0:
                raise ValueError(f'GatherND from opset 11 does not support batch_dims != 0, got {self._batch_dims}')
            return onnx_attrs

        onnx_attrs['batch_dims_i'] = self._batch_dims
        return onnx_attrs

    def forward(self, input_tensor: torch.Tensor, indices: torch.Tensor) -> torch.Tensor:  # pylint: disable=C0116
        def _forward():
            return self._gather_nd(data=input_tensor, indices=indices, batch_dims=self._batch_dims)

        if torch.onnx.is_in_onnx_export():
            onnx_attrs = self._onnx_attrs(opset_version=get_onnx_version())
            return DefaultExportToOnnx.export(_forward, 'GatherND', input_tensor, indices, onnx_attrs)

        return _forward()

    @staticmethod
    def _gather_nd(data: torch.Tensor, indices: torch.Tensor, batch_dims: int) -> torch.Tensor:
        if batch_dims != 0:
            raise NotImplementedError('GatherND for batch_dims != 0 is not implemented')

        r, m = len(data.shape), indices.shape[-1]  # pylint: disable=C0103
        if m > r or m < 1:
            raise ValueError(
                f'The last dimension of indices should have a value between 1 (inclusive) and data rank (inclusive), '
                f'got {m} and {r} respectively'
            )

        total_samples = indices.shape[:-1].numel()
        output_shape = indices.shape[:-1] + data.shape[m:]
        indices_ = torch.split(
            tensor=indices.reshape(total_samples, m).transpose(0, 1),
            split_size_or_sections=1,
            dim=0,
        )

        return data[indices_].reshape(output_shape).contiguous()


@add_converter(operation_type='Gather', version=1)
@add_converter(operation_type='Gather', version=11)
@add_converter(operation_type='Gather', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    torch_module = OnnxGather(
        axis=axis,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='GatherElements', version=11)
@add_converter(operation_type='GatherElements', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    torch_module = OnnxGatherElements(
        axis=axis,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='GatherND', version=11)
@add_converter(operation_type='GatherND', version=12)
@add_converter(operation_type='GatherND', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    batch_dims = node.attributes.get('batch_dims', 0)
    torch_module = OnnxGatherND(
        batch_dims=batch_dims,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



from collections import OrderedDict
from enum import Enum
from types import MappingProxyType
from typing import Mapping
from typing import Tuple

from onnx.onnx_ml_pb2 import GraphProto
from onnx.onnx_ml_pb2 import NodeProto
from onnx.onnx_ml_pb2 import ValueInfoProto

from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.onnx_tensor import OnnxTensor


class ValueType(Enum):  # pylint: disable=missing-class-docstring
    GRAPH_INPUT = 0
    NODE_OUTPUT = 1
    GRAPH_INITIALIZER = 2
    UNKNOWN = 3
    EMPTY = 4


class OnnxGraph:  # pylint: disable=missing-class-docstring
    def __init__(self, onnx_graph_proto: GraphProto):
        self._proto = onnx_graph_proto
        self._input_values = tuple(value_info.name for value_info in self._proto.input)
        self._output_values = tuple(value_info.name for value_info in self._proto.output)

        unique_names = []
        counters = {}
        for node in onnx_graph_proto.node:
            name = OnnxGraph.generate_node_name(node)
            name_counter = counters.setdefault(name, 0)
            counters[name] += 1
            unique_names.append(f'{name}' + (f'_{name_counter}' if name_counter > 0 else ''))

        self._nodes = OrderedDict(
            (name, OnnxNode(node, unique_name=name)) for name, node in zip(unique_names, onnx_graph_proto.node)
        )
        self._initializers = {initializer.name: OnnxTensor(initializer) for initializer in onnx_graph_proto.initializer}
        self._node_output_values = {
            output_name: (node, i) for node in self._nodes.values() for i, output_name in enumerate(node.output_values)
        }
        self._value_info = {value_info.name: value_info for value_info in onnx_graph_proto.value_info}
        for input_value_info in onnx_graph_proto.input:
            self._value_info[input_value_info.name] = input_value_info
        for output_value_info in onnx_graph_proto.output:
            self._value_info[output_value_info.name] = output_value_info

    @property
    def proto(self) -> GraphProto:  # pylint: disable=missing-function-docstring
        return self._proto

    @property
    def value_info(self) -> Mapping[str, ValueInfoProto]:  # pylint: disable=missing-function-docstring
        return self._value_info

    @property
    def name(self) -> str:  # pylint: disable=missing-function-docstring
        return self._proto.name

    @property
    def input_values(self) -> Tuple[str, ...]:  # pylint: disable=missing-function-docstring
        return self._input_values

    @property
    def output_values(self) -> Tuple[str, ...]:  # pylint: disable=missing-function-docstring
        return self._output_values

    @property
    def nodes(self) -> Mapping[str, OnnxNode]:  # pylint: disable=missing-function-docstring
        return self._nodes

    @property
    def initializers(self) -> Mapping[str, OnnxTensor]:  # pylint: disable=missing-function-docstring
        return MappingProxyType(self._initializers)

    def value_type(self, value_name: str) -> ValueType:  # pylint: disable=missing-function-docstring
        if value_name in self._input_values:
            return ValueType.GRAPH_INPUT

        if value_name in self._node_output_values:
            return ValueType.NODE_OUTPUT

        if value_name in self._initializers:
            return ValueType.GRAPH_INITIALIZER

        if value_name == '':
            return ValueType.EMPTY

        return ValueType.UNKNOWN

    def value_as_node_output(  # pylint: disable=missing-function-docstring
        self,
        value_name: str,
    ) -> Tuple[OnnxNode, int]:
        return self._node_output_values[value_name]

    @staticmethod
    def generate_node_name(node: NodeProto) -> str:
        """Generate a torch module name from the given onnx node import it with.

        Uses the ONNX node's name by default, falling back to the op_type in case the former is empty. The node's
        domain is prepended to this.

        Dots (.) are not allowed within names in torch, so they are replaced with a slash (/) instead.

        Parameters
        ----------
        node
            The ONNX node to create a name from.

        Returns
        -------
        A torch-compatible module name based on the given node's properties.
        """
        return (f'{node.domain}/' + (node.name.replace('.', '/') or node.op_type)).lstrip('/')



__all__ = [
    'OnnxResize',
]

import warnings
from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node

_MODES_MAPPING = {
    ('nearest', 1): 'nearest',
    ('nearest', 2): 'nearest',
    ('nearest', 3): 'nearest',
    ('linear', 1): 'linear',
    ('linear', 2): 'bilinear',
    ('linear', 3): 'trilinear',
    ('cubic', 2): 'bicubic',
}


def _get_torch_align_corners(mode: str, coordinate_transformation_mode: str) -> Optional[bool]:
    if mode == 'nearest':
        return None

    return coordinate_transformation_mode == 'align_corners'


def _onnx_mode_to_torch_mode(onnx_mode: str, dim_size: int) -> str:
    torch_mode = _MODES_MAPPING.get((onnx_mode, dim_size), None)
    if torch_mode is None:
        raise NotImplementedError(f'{dim_size}D input is not implemented for "{onnx_mode}" mode.')

    return torch_mode


class OnnxResize(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        mode: str = 'nearest',
        align_corners: Optional[bool] = None,
        ignore_roi: bool = False,
        ignore_bs_ch_size: bool = False,
    ):
        super().__init__()
        self.onnx_mode = mode
        self.align_corners = align_corners
        self.ignore_roi = ignore_roi
        self.ignore_bs_ch_size = ignore_bs_ch_size

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        roi: Optional[torch.Tensor] = None,
        scales: Optional[torch.Tensor] = None,
        sizes: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        torch_mode = _onnx_mode_to_torch_mode(self.onnx_mode, input_tensor.dim() - 2)
        if not self.ignore_roi and roi is not None and roi.nelement() != 0:
            raise NotImplementedError('roi logic is not implemented.')

        # Format of onnx scales and sizes is [n, c, d, h, w]
        # But in torch only [d, h, w] (without batch and channel dimensions)
        if sizes is not None:
            if sizes.nelement() != 0:
                sizes = sizes.tolist()
                input_shape = list(input_tensor.shape)
                if not self.ignore_bs_ch_size and input_shape[:2] != sizes[:2]:
                    raise NotImplementedError('Pytorch\'s interpolate cannot resize channel or batch dimensions.')
                sizes = sizes[2:]
            else:
                sizes = None

        if scales is not None:
            if scales.nelement() != 0:
                scales = scales.tolist()
                if scales[:2] != [1, 1]:
                    raise NotImplementedError('Pytorch\'s interpolate cannot scale channel or batch dimensions.')
                scales = scales[2:]
            else:
                scales = None

        return torch.nn.functional.interpolate(
            input_tensor,
            size=sizes,
            scale_factor=scales,
            mode=torch_mode,
            align_corners=self.align_corners,
        )


class OnnxResizeV10(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, mode: str = 'nearest'):
        super().__init__()
        self._resize = OnnxResize(mode=mode)

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        scales: torch.Tensor,
    ) -> torch.Tensor:
        return self._resize(input_tensor, scales=scales)


@add_converter(operation_type='Resize', version=10)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    mode = node_attributes.get('mode', 'nearest')

    torch_module = OnnxResizeV10(mode=mode)
    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node),
    )


@add_converter(operation_type='Resize', version=11)
@add_converter(operation_type='Resize', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    coordinate_transformation_mode = node_attributes.get('coordinate_transformation_mode', 'half_pixel')
    cubic_coeff_a = node_attributes.get('cubic_coeff_a', -0.75)
    exclude_outside = node_attributes.get('exclude_outside', 0)
    extrapolation_value = node_attributes.get('extrapolation_value', 0.0)
    mode = node_attributes.get('mode', 'nearest')
    nearest_mode = node_attributes.get('nearest_mode', 'round_prefer_floor')

    if mode == 'nearest':
        if nearest_mode != 'floor':
            warnings.warn(
                'Pytorch\'s nearest neighbor interpolate uses the "floor" nearest_mode. '
                'For others modes, the results might differ significantly!'
            )

        if coordinate_transformation_mode != 'asymmetric':
            warnings.warn(
                'Pytorch\'s nearest neighbor interpolation uses "asymmetric" coordinate_transformation_mode. '
                'For others modes, the results might differ significantly!'
            )
    else:
        if coordinate_transformation_mode not in ['pytorch_half_pixel', 'half_pixel']:
            warnings.warn(
                'For linear and cubic interpolation in "asymmetric" and "align_corners" coordinate_transformation_mode'
                'results might differ significantly!'
            )

    if cubic_coeff_a != -0.75:
        warnings.warn('With a cubic coefficient value other than 0.75, the results might differ significantly!')

    if exclude_outside != 0:
        warnings.warn('With a exclude outside value other than 0, the results might differ significantly!')

    if extrapolation_value != 0.0:
        warnings.warn('With a extrapolation value other than 0.0, the results might differ significantly!')

    ignore_roi = coordinate_transformation_mode != 'tf_crop_and_resize'
    return OperationConverterResult(
        torch_module=OnnxResize(
            mode=mode,
            align_corners=_get_torch_align_corners(mode, coordinate_transformation_mode),
            ignore_roi=ignore_roi,
        ),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxRound',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node

_TORCH_ROUND_FROM_ONNX_TYPE = {
    'Ceil': torch.ceil,
    'Floor': torch.floor,
    'Round': torch.round,
}


class OnnxRound(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, round_type: str):
        super().__init__()
        self.round_function = _TORCH_ROUND_FROM_ONNX_TYPE[round_type]

    def forward(self, input_tensor: torch.Tensor):  # pylint: disable=missing-function-docstring
        return self.round_function(input_tensor)


@add_converter(operation_type='Ceil', version=13)
@add_converter(operation_type='Ceil', version=6)
@add_converter(operation_type='Floor', version=13)
@add_converter(operation_type='Floor', version=6)
@add_converter(operation_type='Round', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxRound(node.operation_type),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



# pylint: disable=missing-docstring
__all__ = [
    'OnnxRange',
]

from typing import Union

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxRange(nn.Module, OnnxToTorchModuleWithCustomExport):
    def __init__(self):
        super().__init__()
        self.register_buffer('dummy_buffer', torch.Tensor(), persistent=False)

    @staticmethod
    def _get_scalar(value) -> Union[float, int]:
        if isinstance(value, torch.Tensor):
            return value.item()

        return value

    def _arange(
        self,
        start: Union[torch.Tensor, float, int],
        limit: Union[torch.Tensor, float, int],
        delta: Union[torch.Tensor, float, int],
    ) -> torch.Tensor:
        return torch.arange(
            start=self._get_scalar(start),
            end=self._get_scalar(limit),
            step=self._get_scalar(delta),
            device=self.dummy_buffer.device,
        )

    def forward(
        self,
        start: Union[torch.Tensor, float, int],
        limit: Union[torch.Tensor, float, int],
        delta: Union[torch.Tensor, float, int],
    ) -> torch.Tensor:
        def _forward() -> torch.Tensor:
            return self._arange(start, limit, delta)

        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(_forward, 'Range', start, limit, delta, {})

        return _forward()


@add_converter(operation_type='Range', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    return OperationConverterResult(
        torch_module=OnnxRange(),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxCompare',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node

_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'Equal': torch.eq,
    'Less': torch.less,
    'LessOrEqual': torch.less_equal,
    'Greater': torch.greater,
    'GreaterOrEqual': torch.greater_equal,
}


class OnnxCompare(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, operation_type: str):
        super().__init__()
        self.compare_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return self.compare_function(x, y)


@add_converter(operation_type='Equal', version=7)
@add_converter(operation_type='Equal', version=11)
@add_converter(operation_type='Equal', version=13)
@add_converter(operation_type='Less', version=7)
@add_converter(operation_type='Less', version=9)
@add_converter(operation_type='Less', version=13)
@add_converter(operation_type='Greater', version=7)
@add_converter(operation_type='Greater', version=9)
@add_converter(operation_type='Greater', version=13)
@add_converter(operation_type='LessOrEqual', version=12)
@add_converter(operation_type='GreaterOrEqual', version=12)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxCompare(operation_type=node.operation_type),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = []

from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_shape_from_value_info
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.padding import onnx_auto_pad_to_torch_padding

_AVGPOOL_CLASS_FROM_SPATIAL_RANK = {
    1: nn.AvgPool1d,
    2: nn.AvgPool2d,
    3: nn.AvgPool3d,
}


@add_converter(operation_type='AveragePool', version=7)
@add_converter(operation_type='AveragePool', version=10)
@add_converter(operation_type='AveragePool', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    input_value_info = graph.value_info[node.input_values[0]]
    input_shape = get_shape_from_value_info(input_value_info)

    spatial_rank = len(input_shape) - 2
    try:
        avgpool_class = _AVGPOOL_CLASS_FROM_SPATIAL_RANK[spatial_rank]
    except KeyError as exc:
        raise NotImplementedError(
            f'Average pool operation with spatial rank == {spatial_rank} is not implemented'
        ) from exc

    node_attributes = node.attributes
    # required
    kernel_shape = node_attributes['kernel_shape']
    # optional
    ceil_mode = node_attributes.get('ceil_mode', 0)
    strides = node_attributes.get('strides', 1)
    count_include_pad = node_attributes.get('count_include_pad', 0)

    padding, padding_module = onnx_auto_pad_to_torch_padding(
        onnx_padding=node_attributes.get('pads', [0] * spatial_rank * 2),
        auto_pad=node_attributes.get('auto_pad', 'NOTSET'),
    )
    if padding_module is not None:
        raise NotImplementedError('AvgPool with non symmetrical padding is not implemented.')

    torch_module = avgpool_class(
        kernel_size=kernel_shape,
        stride=strides,
        padding=padding,
        count_include_pad=count_include_pad == 1,
        ceil_mode=ceil_mode == 1,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxConstantOfShape',
]

from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxConstantOfShape(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, value: Optional[torch.Tensor] = None):
        super().__init__()

        if value is None:
            value = torch.tensor(0.0, dtype=torch.float32)

        if value.numel() != 1:
            raise ValueError('parameter "value" must be scalar')

        self.value: torch.Tensor
        self.register_buffer('value', value)

    def forward(self, shape: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        fill_value = self.value.item()

        return torch.full(
            size=torch.Size(shape),
            fill_value=int(fill_value) if isinstance(fill_value, bool) else fill_value,
            dtype=self.value.dtype,
            device=self.value.device,
        )


@add_converter(operation_type='ConstantOfShape', version=9)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes

    if 'value' in node_attributes:
        value = node_attributes['value'].to_torch()
    else:
        value = None

    return OperationConverterResult(
        torch_module=OnnxConstantOfShape(value=value),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxConcat',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxConcat(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, axis: int):
        super().__init__()
        self.axis = axis

    def forward(self, *input_tensors) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return torch.cat(input_tensors, self.axis)


@add_converter(operation_type='Concat', version=4)
@add_converter(operation_type='Concat', version=11)
@add_converter(operation_type='Concat', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    torch_module = OnnxConcat(
        axis=axis,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



import logging
from typing import Callable
from typing import NamedTuple

from onnx import defs

from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult

_LOGGER = logging.getLogger(__name__)
_CONVERTER_REGISTRY = {}


class OperationDescription(NamedTuple):  # pylint: disable=missing-class-docstring
    domain: str
    operation_type: str
    version: int


TConverter = Callable[[OnnxNode, OnnxGraph], OperationConverterResult]


def add_converter(  # pylint: disable=missing-function-docstring
    operation_type: str,
    version: int,
    domain: str = defs.ONNX_DOMAIN,
    override_registry: bool = False,
):
    description = OperationDescription(
        domain=domain,
        operation_type=operation_type,
        version=version,
    )

    def deco(converter: TConverter):
        if description in _CONVERTER_REGISTRY and not override_registry:
            raise ValueError(f'Operation "{description}" already registered')

        _CONVERTER_REGISTRY[description] = converter
        _LOGGER.debug(f'Operation converter registered {description}')

        return converter

    return deco


def get_converter(  # pylint: disable=missing-function-docstring
    operation_type: str,
    version: int,
    domain: str = defs.ONNX_DOMAIN,
) -> TConverter:
    try:
        version = defs.get_schema(
            operation_type,
            domain=domain,
            max_inclusive_version=version,
        ).since_version
    except (RuntimeError, defs.SchemaError):
        pass

    description = OperationDescription(
        domain=domain,
        operation_type=operation_type,
        version=version,
    )

    converter = _CONVERTER_REGISTRY.get(description, None)
    if converter is None:
        raise NotImplementedError(f'Converter is not implemented ({description})')

    return converter



import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxCopyIdentity(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def forward(self, x: torch.Tensor):  # pylint: disable=missing-function-docstring
        return x.clone()


@add_converter(operation_type='Identity', version=16)
@add_converter(operation_type='Identity', version=14)
@add_converter(operation_type='Identity', version=13)
@add_converter(operation_type='Identity', version=1)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    # We need copy identity because in onnx identity create new tensor.
    # Pytorch identity simply returns the same tensor.
    # Which ruin quantization logic, because we should mark quantized tensors.
    # For example, input quantization node will be supressed if input tensor is already quantized.
    return OperationConverterResult(
        torch_module=OnnxCopyIdentity(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxExpand',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxExpand(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-docstring
    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        shape: torch.Tensor,
    ) -> torch.Tensor:
        if shape.is_cuda:
            shape = shape.cpu()

        def _forward():
            return input_tensor * torch.ones(
                torch.Size(shape.tolist()), dtype=input_tensor.dtype, device=input_tensor.device
            )

        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(_forward, 'Expand', input_tensor, shape, {})

        return _forward()


@add_converter(operation_type='Expand', version=8)
@add_converter(operation_type='Expand', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxExpand(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



from typing import Any
from typing import Dict
from typing import Type
from typing import Union

import numpy as np
import torch


def onnx_dtype_to_torch_dtype(dtype: int) -> Union[torch.dtype, Type[str], Type[bool]]:
    """
    Convert ONNX dtype to PyTorch dtype.

    Parameters
    ----------
    dtype : int
        ONNX data type.

    Returns
    -------
    Union[torch.dtype, Type[str], Type[bool]]
        Corresponding PyTorch dtype.

    """
    # https://github.com/onnx/onnx/blob/main/onnx/onnx-ml.proto#L485
    _dtypes: Dict[int, Union[torch.dtype, Type[str], Type[bool]]] = {
        1: torch.float32,
        2: torch.uint8,
        3: torch.int8,
        # 4: UINT16 is not supported: https://github.com/pytorch/pytorch/issues/58734.
        5: torch.int16,
        6: torch.int32,
        7: torch.int64,
        8: str,
        9: bool,
        10: torch.float16,
        11: torch.float64,
        # 12: UINT32 is not supported: https://github.com/pytorch/pytorch/issues/58734.
        # 13: UINT64 is not supported: https://github.com/pytorch/pytorch/issues/58734.
        14: torch.complex64,
        15: torch.complex128,
        16: torch.bfloat16,
    }
    try:
        return _dtypes[dtype]
    except KeyError as exc:
        raise ValueError(f'dtype={dtype} is not supported') from exc


def onnx_dtype_to_numpy_dtype(dtype: int) -> Union[np.dtype, Type[str], Type[bool]]:
    """
    Convert ONNX dtype to Numpy dtype.

    Parameters
    ----------
    dtype : int
        ONNX data type.

    Returns
    -------
    Union[torch.dtype, Type[str], Type[bool]]
        Corresponding Numpy dtype.

    """
    # https://numpy.org/doc/stable/reference/arrays.dtypes.html
    _dtypes: Dict[int, Any] = {
        1: np.float32,
        2: np.uint8,
        3: np.int8,
        4: np.uint16,
        5: np.int16,
        6: np.int32,
        7: np.int64,
        8: str,
        9: bool,
        10: np.float16,
        11: np.float64,
        12: np.uint32,
        13: np.uint64,
        14: np.complex64,
        15: np.complex128,
        # 16: bfloat16 is not supported.
    }
    try:
        return _dtypes[dtype]
    except KeyError as exc:
        raise ValueError(f'dtype={dtype} is not supported') from exc



__all__ = [
    'OnnxEyeLike',
]

from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.dtype import onnx_dtype_to_torch_dtype


class OnnxEyeLike(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, dtype: Optional[int] = None, k: int = 0):  # pylint: disable=invalid-name
        super().__init__()
        self.dtype = dtype
        self.k = k  # pylint: disable=invalid-name

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if len(x.shape) != 2:
            raise ValueError(f'EyeLike only supports 2D tensors, got {len(x.shape)}')

        dtype = x.dtype if self.dtype is None else onnx_dtype_to_torch_dtype(self.dtype)
        if not isinstance(dtype, torch.dtype):
            raise ValueError(f'Expected type of dtype is torch.dtype, got {type(dtype)}')

        rows, cols = x.size()
        if self.k > rows:
            raise ValueError(
                f'EyeLike attribute k should be less or equal than the zero dimension of input tensor,'
                f'got {self.k} and {rows}'
            )

        if self.k == 0:
            return torch.eye(n=rows, m=cols, dtype=dtype)
        if self.k > 0:
            return torch.concat(
                [
                    torch.zeros(rows, self.k, dtype=dtype),
                    torch.eye(n=rows, m=(cols - self.k), dtype=dtype),
                ],
                dim=1,
            )
        return torch.concat(  # k < 0:
            [
                torch.zeros(-self.k, cols, dtype=dtype),
                torch.eye(n=(rows + self.k), m=cols, dtype=dtype),
            ],
            dim=0,
        )


@add_converter(operation_type='EyeLike', version=9)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    k = node_attributes.get('k', 0)  # pylint: disable=invalid-name
    dtype = node_attributes.get('dtype', None)
    return OperationConverterResult(
        torch_module=OnnxEyeLike(dtype=dtype, k=k),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxClip',
]

from typing import Optional

import torch
from torch import nn
from torch.types import Number

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_const_value
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxClip(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(
        self,
        min_val: Optional[Number] = None,
        max_val: Optional[Number] = None,
    ):
        super().__init__()
        self.min_val = min_val
        self.max_val = max_val

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return torch.clamp(input_tensor, self.min_val, self.max_val)


def _create_torch_module(min_val: Optional[torch.Tensor], max_val: Optional[torch.Tensor]) -> nn.Module:
    if min_val is None and max_val is None:
        torch_module = nn.Identity()
    elif min_val == 0 and max_val is None:
        torch_module = nn.ReLU()
    elif min_val == 0 and max_val == 6:
        torch_module = nn.ReLU6()
    else:
        torch_module = OnnxClip(min_val=min_val, max_val=max_val)

    return torch_module


@add_converter(operation_type='Clip', version=11)
@add_converter(operation_type='Clip', version=12)
@add_converter(operation_type='Clip', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    # Min and Max inputs are optional
    min_name = node.input_values[1] if len(node.input_values) > 1 else None
    max_name = node.input_values[2] if len(node.input_values) > 2 else None

    try:
        min_val = float(get_const_value(min_name, graph)) if min_name is not None and min_name != "" else None
        max_val = float(get_const_value(max_name, graph)) if max_name is not None and max_name != "" else None
    except KeyError as exc:
        raise NotImplementedError('Dynamic value of min/max is not implemented') from exc

    torch_module = _create_torch_module(min_val=min_val, max_val=max_val)

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=OnnxMapping(
            inputs=(node.input_values[0],),
            outputs=node.output_values,
        ),
    )


@add_converter(operation_type='Clip', version=6)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    min_val = node_attributes.get('min', None)
    max_val = node_attributes.get('max', None)

    torch_module = _create_torch_module(min_val=min_val, max_val=max_val)

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxMean',
]

import torch

from onnx2torch2.node_converters.base_element_wise import OnnxBaseElementWise
from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxMean(OnnxBaseElementWise):  # pylint: disable=missing-docstring
    def __init__(self):
        super().__init__(op_type='Mean')

    def apply_reduction(self, *tensors: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        broadcast_shape = self._broadcast_shape(*tensors)

        output = torch.zeros(broadcast_shape, dtype=tensors[0].dtype, device=tensors[0].device)
        for y in tensors:
            output.add_(y)

        output = output.div(len(tensors))  # Divide by the number of tensors
        return output


@add_converter(operation_type='Mean', version=8)
@add_converter(operation_type='Mean', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxMean(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



# pylint: disable=missing-docstring
__all__ = [
    'OnnxNonMaxSuppression',
]

from typing import Any
from typing import Dict
from typing import Optional

import torch
import torchvision
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxNonMaxSuppression(nn.Module, OnnxToTorchModuleWithCustomExport):
    def __init__(self, center_point_box: int = 0):
        super().__init__()
        self._center_point_box = center_point_box

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        del opset_version
        return {'center_point_box_i': self._center_point_box}

    def forward(
        self,
        boxes: torch.Tensor,
        scores: torch.Tensor,
        max_output_boxes_per_class: Optional[torch.Tensor] = None,
        iou_threshold: Optional[torch.Tensor] = None,
        score_threshold: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        def _forward() -> torch.Tensor:
            return self._nms(boxes, scores, max_output_boxes_per_class, iou_threshold, score_threshold)

        if torch.onnx.is_in_onnx_export():
            if max_output_boxes_per_class is None:
                max_output_boxes_per_class = torch.tensor([0], dtype=torch.int64)
            if iou_threshold is None:
                iou_threshold = torch.tensor([0.0], dtype=torch.float32)
            if score_threshold is None:
                score_threshold = torch.tensor([0.0], dtype=torch.float32)

            onnx_attrs = self._onnx_attrs(opset_version=get_onnx_version())
            return DefaultExportToOnnx.export(
                _forward,
                'NonMaxSuppression',
                boxes,
                scores,
                max_output_boxes_per_class,
                iou_threshold,
                score_threshold,
                onnx_attrs,
            )

        return _forward()

    def _nms(
        self,
        boxes: torch.Tensor,
        scores: torch.Tensor,
        max_output_boxes_per_class: Optional[torch.Tensor],
        iou_threshold: Optional[torch.Tensor],
        score_threshold: Optional[torch.Tensor],
    ) -> torch.Tensor:
        if max_output_boxes_per_class is None:
            return torch.empty([0, 3], dtype=torch.int64, device=boxes.device)

        max_output_boxes_per_class = max_output_boxes_per_class.item()
        iou_threshold = 0.0 if iou_threshold is None else iou_threshold.item()
        score_threshold = 0.0 if score_threshold is None else score_threshold.item()

        out = []
        # boxes - [bs, num_boxes, 4], scores - [bs, n_classes, num_boxes]
        for batch_index, (batch_boxes, batch_scores) in enumerate(zip(boxes, scores)):
            # bbox - [num_boxes, 4], score - [n_classes, num_boxes]
            for class_index, class_scores in enumerate(batch_scores):
                confidence_mask = class_scores > score_threshold
                confidence_indexes = confidence_mask.nonzero(as_tuple=False).squeeze(1)

                filtered_batch_boxes = batch_boxes[confidence_indexes]
                if self._center_point_box:
                    filtered_batch_boxes = torchvision.ops.box_convert(
                        filtered_batch_boxes,
                        in_fmt='cxcywh',
                        out_fmt='xyxy',
                    )

                nms_indexes = torchvision.ops.nms(
                    boxes=filtered_batch_boxes,
                    scores=class_scores[confidence_indexes],
                    iou_threshold=iou_threshold,
                )
                num_boxes = min(max_output_boxes_per_class, nms_indexes.size(0))
                nms_indexes = nms_indexes[:num_boxes]
                indexes = confidence_indexes[nms_indexes]

                out.extend([batch_index, class_index, box_index] for box_index in indexes)
        if len(out) == 0:
            return torch.empty([0, 3], dtype=torch.int64, device=boxes.device)

        return torch.tensor(out, dtype=torch.int64, device=boxes.device)


@add_converter(operation_type='NonMaxSuppression', version=10)
@add_converter(operation_type='NonMaxSuppression', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    center_point_box = node.attributes.get('center_point_box', 0)
    return OperationConverterResult(
        torch_module=OnnxNonMaxSuppression(center_point_box=center_point_box),
        onnx_mapping=onnx_mapping_from_node(node),
    )



from types import MappingProxyType
from typing import Any
from typing import Mapping
from typing import Tuple

from onnx.onnx_ml_pb2 import AttributeProto
from onnx.onnx_ml_pb2 import NodeProto

from onnx2torch2.onnx_tensor import OnnxTensor


class OnnxNode:  # pylint: disable=missing-class-docstring
    def __init__(self, onnx_node_proto: NodeProto, unique_name: str):
        self._proto = onnx_node_proto
        self._unique_name = unique_name
        self._input_values = tuple(onnx_node_proto.input)
        self._output_values = tuple(onnx_node_proto.output)
        self._inputs = None

        self._proto_attributes = {
            attribute.name: OnnxNode._parse_attribute_value(attribute) for attribute in self._proto.attribute
        }

    @staticmethod
    def _parse_attribute_value(attribute: AttributeProto) -> Any:
        if attribute.HasField('i'):
            value = attribute.i
        elif attribute.HasField('f'):
            value = attribute.f
        elif attribute.HasField('s'):
            value = str(attribute.s, 'utf-8')
        elif attribute.HasField('t'):
            value = OnnxTensor(attribute.t)
        elif attribute.ints:
            value = list(attribute.ints)
        elif attribute.floats:
            value = list(attribute.floats)
        elif attribute.strings:
            value = [str(s, 'utf-8') for s in attribute.strings]
        elif attribute.tensors:
            value = [OnnxTensor(t) for t in attribute.tensors]
        else:
            value = attribute

        return value

    @property
    def proto(self) -> NodeProto:  # pylint: disable=missing-function-docstring
        return self._proto

    @property
    def name(self) -> str:  # pylint: disable=missing-function-docstring
        return self._proto.name

    @property
    def unique_name(self) -> str:  # pylint: disable=missing-function-docstring
        return self._unique_name

    @property
    def domain(self) -> str:  # pylint: disable=missing-function-docstring
        return self._proto.domain

    @property
    def operation_type(self) -> str:  # pylint: disable=missing-function-docstring
        return self._proto.op_type

    @property
    def input_values(self) -> Tuple[str, ...]:  # pylint: disable=missing-function-docstring
        return self._input_values

    @property
    def output_values(self) -> Tuple[str, ...]:  # pylint: disable=missing-function-docstring
        return self._output_values

    @property
    def attributes(self) -> Mapping[str, Any]:  # pylint: disable=missing-function-docstring
        return MappingProxyType(self._proto_attributes)



__all__ = [
    'OnnxCast',
]

import torch
from onnx import TensorProto  # pylint: disable=no-name-in-module
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node

# pylint: disable=no-member
TENSOR_TYPE_TO_TORCH_TYPE = {
    int(TensorProto.FLOAT): torch.float32,
    int(TensorProto.UINT8): torch.uint8,
    int(TensorProto.INT8): torch.int8,
    int(TensorProto.INT16): torch.int16,
    int(TensorProto.INT32): torch.int32,
    int(TensorProto.INT64): torch.int64,
    int(TensorProto.BOOL): torch.bool,
    int(TensorProto.FLOAT16): torch.float16,
    int(TensorProto.DOUBLE): torch.float64,
    int(TensorProto.COMPLEX64): torch.complex64,
    int(TensorProto.COMPLEX128): torch.complex128,
}
# pylint: enable=no-member


class OnnxCast(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, onnx_dtype: int):
        super().__init__()
        try:
            self.torch_dtype = TENSOR_TYPE_TO_TORCH_TYPE[onnx_dtype]
        except KeyError as exc:
            raise NotImplementedError(f'Conversion to "{onnx_dtype}" is not implemented') from exc

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return input_tensor.to(self.torch_dtype)


@add_converter(operation_type='Cast', version=9)
@add_converter(operation_type='Cast', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    onnx_dtype = node_attributes.get('to', None)

    return OperationConverterResult(
        torch_module=OnnxCast(onnx_dtype),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



# pylint: disable=missing-docstring
__all__ = [
    'OnnxNot',
    'OnnxLogical',
]

from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import old_style_broadcast
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport

_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'Or': torch.logical_or,
    'And': torch.logical_and,
    'Xor': torch.logical_xor,
}


class OnnxNot(nn.Module, OnnxToTorchModuleWithCustomExport):
    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:
        def _forward() -> torch.Tensor:
            return torch.logical_not(input_tensor)

        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(_forward, 'Not', input_tensor, {})

        return _forward()


class OnnxLogical(nn.Module, OnnxToTorchModule):
    def __init__(self, operation_type: str, broadcast: Optional[int] = None, axis: Optional[int] = None):
        super().__init__()
        self.broadcast = broadcast
        self.axis = axis

        self.logic_op_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

    def forward(self, first_tensor: torch.Tensor, second_tensor: torch.Tensor):
        if self.broadcast == 1 and self.axis is not None:
            second_tensor = old_style_broadcast(first_tensor, second_tensor, self.axis)

        return self.logic_op_function(first_tensor, second_tensor)


@add_converter(operation_type='Xor', version=1)
@add_converter(operation_type='Xor', version=7)
@add_converter(operation_type='And', version=1)
@add_converter(operation_type='And', version=7)
@add_converter(operation_type='Or', version=1)
@add_converter(operation_type='Or', version=7)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    return OperationConverterResult(
        torch_module=OnnxLogical(
            operation_type=node.operation_type,
            broadcast=node.attributes.get('broadcast', None),
            axis=node.attributes.get('axis', None),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='Not', version=1)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    return OperationConverterResult(
        torch_module=OnnxNot(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxBinaryMathOperation',
]

from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import old_style_broadcast
from onnx2torch2.utils.common import onnx_mapping_from_node


def _onnx_div(first: torch.Tensor, second: torch.Tensor) -> torch.Tensor:
    if first.is_floating_point() or second.is_floating_point():  # float division
        return torch.div(first, second)

    return torch.div(first, second, rounding_mode='trunc')  # integer division


_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'Add': torch.add,
    'Sub': torch.sub,
    'Mul': torch.mul,
    'Div': _onnx_div,
}


class OnnxBinaryMathOperation(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, operation_type: str, broadcast: Optional[int] = None, axis: Optional[int] = None):
        super().__init__()

        self.broadcast = broadcast
        self.axis = axis
        self.math_op_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

    def forward(  # pylint: disable=missing-function-docstring
        self,
        first: torch.Tensor,
        second: torch.Tensor,
    ) -> torch.Tensor:
        if self.broadcast == 1 and self.axis is not None:
            second = old_style_broadcast(first, second, self.axis)

        return self.math_op_function(first, second)


@add_converter(operation_type='Add', version=1)
@add_converter(operation_type='Add', version=6)
@add_converter(operation_type='Add', version=7)
@add_converter(operation_type='Add', version=13)
@add_converter(operation_type='Add', version=14)
@add_converter(operation_type='Sub', version=1)
@add_converter(operation_type='Sub', version=6)
@add_converter(operation_type='Sub', version=7)
@add_converter(operation_type='Sub', version=13)
@add_converter(operation_type='Sub', version=14)
@add_converter(operation_type='Mul', version=1)
@add_converter(operation_type='Mul', version=6)
@add_converter(operation_type='Mul', version=7)
@add_converter(operation_type='Mul', version=13)
@add_converter(operation_type='Mul', version=14)
@add_converter(operation_type='Div', version=1)
@add_converter(operation_type='Div', version=6)
@add_converter(operation_type='Div', version=7)
@add_converter(operation_type='Div', version=13)
@add_converter(operation_type='Div', version=14)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxBinaryMathOperation(
            operation_type=node.operation_type,
            broadcast=node.attributes.get('broadcast', None),
            axis=node.attributes.get('axis', None),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxCumSum',
]
import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


def _arbitrary_dim_shift_and_insert_zero(
    input_tensor: torch.Tensor,
    insert_dim: int,
) -> torch.Tensor:
    # single item shift
    slice_index, insertion = [[slice(None)] * len(input_tensor.shape)] * 2
    insert_dim_size = input_tensor.shape[insert_dim]

    slice_index[insert_dim] = slice(0, -1)
    slice_index = tuple(slice_index)
    tensor_slice = input_tensor[slice_index]

    insert_index = torch.arange(start=1, end=insert_dim_size, dtype=torch.int64, device=input_tensor.device)
    index_shape = [1] * len(input_tensor.shape)
    index_shape[insert_dim] = insert_dim_size - 1

    insert_index = torch.reshape(insert_index, index_shape)
    insert_index = insert_index + torch.zeros_like(tensor_slice, dtype=torch.int64, device=input_tensor.device)

    input_tensor = torch.scatter(
        input=input_tensor,
        dim=insert_dim,
        index=insert_index,
        src=tensor_slice,
    )

    insertion[insert_dim] = slice(0, 1)
    insertion = tuple(insertion)
    input_tensor[insertion] = 0

    return input_tensor


class OnnxCumSum(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(
        self,
        exclusive: bool = False,
        reverse: bool = False,
    ):
        super().__init__()
        self.exclusive = exclusive
        self.reverse = reverse

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        axis: torch.Tensor,
    ) -> torch.Tensor:
        axis = axis.item()
        if self.reverse:
            input_tensor = torch.flip(input_tensor, dims=(axis,))

        if self.exclusive:
            input_tensor = _arbitrary_dim_shift_and_insert_zero(input_tensor, insert_dim=axis)

        input_tensor = torch.cumsum(input_tensor, dim=axis)

        if self.reverse:
            input_tensor = torch.flip(input_tensor, dims=(axis,))

        return input_tensor


@add_converter(operation_type='CumSum', version=11)
@add_converter(operation_type='CumSum', version=14)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    exclusive = bool(node_attributes.get('exclusive', 0))
    reverse = bool(node_attributes.get('reverse', 1))

    return OperationConverterResult(
        torch_module=OnnxCumSum(exclusive, reverse),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxUnsqueezeStaticAxes',
    'OnnxUnsqueezeDynamicAxes',
]

from typing import List

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_const_value
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxUnsqueezeStaticAxes(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, axes: List[int]):
        super().__init__()
        self._axes = sorted(axes)

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        result = input_tensor
        for axes_id in self._axes:
            result = torch.unsqueeze(result, dim=axes_id)

        return result


class OnnxUnsqueezeDynamicAxes(  # pylint: disable=missing-class-docstring
    nn.Module,
    OnnxToTorchModuleWithCustomExport,
):
    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        axes: torch.Tensor,
    ) -> torch.Tensor:
        def _forward():
            result = input_tensor
            for axes_id in torch.sort(axes).values:
                result = torch.unsqueeze(result, dim=axes_id)

            return result

        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(_forward, 'Unsqueeze', input_tensor, axes, {})

        return _forward()


@add_converter(operation_type='Unsqueeze', version=1)
@add_converter(operation_type='Unsqueeze', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axes = node.attributes['axes']
    return OperationConverterResult(
        torch_module=OnnxUnsqueezeStaticAxes(axes=axes),
        onnx_mapping=onnx_mapping_from_node(node),
    )


@add_converter(operation_type='Unsqueeze', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    try:
        axes = get_const_value(node.input_values[1], graph)
        axes = axes.tolist()
        return OperationConverterResult(
            torch_module=OnnxUnsqueezeStaticAxes(axes=axes),
            onnx_mapping=OnnxMapping(
                inputs=(node.input_values[0],),
                outputs=node.output_values,
            ),
        )
    except KeyError:
        pass

    return OperationConverterResult(
        torch_module=OnnxUnsqueezeDynamicAxes(),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxSplit',
    'OnnxSplit13',
]

from typing import List
from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxSplit13(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, num_splits: int, axis: int = 0):
        super().__init__()

        self.axis = axis
        self.num_splits = num_splits

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        split: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if split is None:
            axis_len = input_tensor.shape[self.axis]
            split_size_or_sections = axis_len // self.num_splits
        else:
            split_size_or_sections = split.tolist()

        return torch.split(input_tensor, split_size_or_sections, dim=self.axis)


class OnnxSplit(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, num_splits: int, axis: int = 0, split: Optional[List[int]] = None):
        super().__init__()

        self.axis = axis
        self.num_splits = num_splits
        self.split = split

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self.split is None:
            axis_len = input_tensor.shape[self.axis]
            split_size_or_sections = axis_len // self.num_splits
        else:
            split_size_or_sections = self.split

        return torch.split(input_tensor, split_size_or_sections, dim=self.axis)


@add_converter(operation_type='Split', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    num_splits = len(node.output_values)
    return OperationConverterResult(
        torch_module=OnnxSplit13(axis=axis, num_splits=num_splits),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='Split', version=11)
@add_converter(operation_type='Split', version=2)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    split = node.attributes.get('split', None)
    num_splits = len(node.output_values)
    return OperationConverterResult(
        torch_module=OnnxSplit(axis=axis, split=split, num_splits=num_splits),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = ['OnnxDepthToSpace']

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxDepthToSpace(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, blocksize: int, mode: str):
        super().__init__()
        self.blocksize = blocksize
        self.mode = mode

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self.mode == "DCR":
            b, c, h, w = input_tensor.shape
            tmp = torch.reshape(input_tensor, [b, self.blocksize, self.blocksize, c // (self.blocksize**2), h, w])
            tmp = torch.permute(tmp, [0, 3, 4, 1, 5, 2])
            return torch.reshape(tmp, [b, c // (self.blocksize**2), h * self.blocksize, w * self.blocksize])
        return torch.pixel_shuffle(input_tensor, upscale_factor=self.blocksize)


@add_converter(operation_type='DepthToSpace', version=11)
@add_converter(operation_type='DepthToSpace', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph

    blocksize: int = node.attributes['blocksize']  # required
    mode: str = node.attributes.get('mode', 'DCR')

    return OperationConverterResult(
        torch_module=OnnxDepthToSpace(blocksize=blocksize, mode=mode),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxShape',
]

from typing import Any
from typing import Dict
from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxShape(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(self, start: int = 0, end: Optional[int] = None):
        super().__init__()
        self._start = start
        self._end = end

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        if opset_version < 15:
            if self._start != 0:
                raise ValueError(f'Shape from opset < 15 does not support start != 0, got {self._start}')
            if self._end is not None:
                raise ValueError(f'Shape from opset < 15 does not support end != None, got {self._end}')
            return {}

        onnx_attrs: Dict[str, Any] = {'start_i': self._start}
        if self._end:
            onnx_attrs['end_i'] = self._end

        return onnx_attrs

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        def _forward():
            return torch.tensor(
                input_tensor.shape[self._start : self._end],
                device=input_tensor.device,
            )

        if torch.onnx.is_in_onnx_export():
            onnx_attrs = self._onnx_attrs(opset_version=get_onnx_version())
            return DefaultExportToOnnx.export(_forward, 'Shape', input_tensor, onnx_attrs)

        return _forward()


@add_converter(operation_type='Shape', version=1)
@add_converter(operation_type='Shape', version=13)
@add_converter(operation_type='Shape', version=15)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxShape(
            start=node.attributes.get('start', 0),
            end=node.attributes.get('end', None),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxMatMul',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxMatMul(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, name):
        super().__init__()
        self.name = name

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        # TODO: Figure out a way to simplify this code
        if x.dim() == 3 and y.dim() == 3 and x.shape[0] == y.shape[0]:
            out = torch.bmm(x, y)
        else:
            try:
                out = torch.matmul(x, y)
            except:  # pylint: disable=bare-except
                out = torch.mm(x, y)
        return out


@add_converter(operation_type='MatMul', version=1)
@add_converter(operation_type='MatMul', version=9)
@add_converter(operation_type='MatMul', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxMatMul(node.name),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxSqueezeStaticAxes',
    'OnnxSqueezeDynamicAxes',
]

from typing import List
from typing import Optional

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxSqueezeStaticAxes(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(self, axes: Optional[List[int]] = None):
        super().__init__()
        if axes is not None:
            axes = sorted(axes, reverse=True)

        self.axes = axes

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        def _forward():
            if not self.axes:
                return torch.squeeze(input_tensor)

            result = input_tensor
            for axes_id in self.axes:
                result = torch.squeeze(result, dim=axes_id)

            return result

        if torch.onnx.is_in_onnx_export() and get_onnx_version() >= 13:
            args = [input_tensor]
            if self.axes:
                axes = torch.tensor(self.axes, device=input_tensor.device, dtype=torch.int64)
                args.append(axes)

            return DefaultExportToOnnx.export(_forward, 'Squeeze', *args, {})

        return _forward()


class OnnxSqueezeDynamicAxes(  # pylint: disable=missing-class-docstring
    nn.Module,
    OnnxToTorchModuleWithCustomExport,
):
    @staticmethod
    def is_empty_axes(axes: torch.Tensor) -> bool:  # pylint: disable=missing-function-docstring
        return axes is None or axes.nelement() == 0

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        axes: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        def _forward():
            if OnnxSqueezeDynamicAxes.is_empty_axes(axes):
                return torch.squeeze(input_tensor)

            result = input_tensor
            for axes_id in torch.sort(axes, descending=True).values:
                result = torch.squeeze(result, dim=axes_id)

            return result

        if torch.onnx.is_in_onnx_export():
            args = [input_tensor]
            if not self.is_empty_axes(axes):
                args.append(axes)

            return DefaultExportToOnnx.export(_forward, 'Squeeze', *args, {})

        return _forward()


@add_converter(operation_type='Squeeze', version=1)
@add_converter(operation_type='Squeeze', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axes = node.attributes.get('axes', None)
    return OperationConverterResult(
        torch_module=OnnxSqueezeStaticAxes(axes=axes),
        onnx_mapping=onnx_mapping_from_node(node),
    )


@add_converter(operation_type='Squeeze', version=13)
@add_converter(operation_type='Squeeze', version=21)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxSqueezeDynamicAxes(),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxInstanceNorm',
]

import torch
import torch.nn.functional as F
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_shape_from_value_info
from onnx2torch2.utils.common import onnx_mapping_from_node

_IN_CLASS_FROM_SPATIAL_RANK = {
    0: nn.InstanceNorm1d,
    1: nn.InstanceNorm1d,
    2: nn.InstanceNorm2d,
    3: nn.InstanceNorm3d,
}


class OnnxInstanceNorm(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, momentum: float, epsilon: float):
        super().__init__()
        self.momentum = momentum
        self.epsilon = epsilon

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_data: torch.Tensor,
        weight: torch.Tensor,
        bias: torch.Tensor,
    ) -> torch.Tensor:
        return F.instance_norm(
            input=input_data,
            running_mean=None,
            running_var=None,
            weight=weight,
            bias=bias,
            use_input_stats=True,
            momentum=self.momentum,
            eps=self.epsilon,
        )


@add_converter(operation_type='InstanceNormalization', version=1)
@add_converter(operation_type='InstanceNormalization', version=6)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    node_attributes = node.attributes
    epsilon = node_attributes.get('epsilon', 1e-5)
    momentum = 0.1

    if all(value_name in graph.initializers for value_name in node.input_values[1:]):
        input_value_info = graph.value_info[node.input_values[0]]
        input_shape = get_shape_from_value_info(input_value_info)
        spatial_rank = len(input_shape) - 2
        try:
            in_class = _IN_CLASS_FROM_SPATIAL_RANK[spatial_rank]
        except KeyError as exc:
            raise NotImplementedError(
                f'InstanceNorm operation with spatial rank == {spatial_rank} is not implemented'
            ) from exc

        scale_value_name = node.input_values[1]
        bias_value_name = node.input_values[2]

        scale = graph.initializers[scale_value_name].to_torch()
        torch_module = in_class(
            num_features=scale.size()[0],
            eps=epsilon,
            momentum=momentum,
            affine=True,
            track_running_stats=False,
        )
        with torch.no_grad():
            torch_module.weight.data = graph.initializers[scale_value_name].to_torch()
            torch_module.bias.data = graph.initializers[bias_value_name].to_torch()

        onnx_mapping = OnnxMapping(inputs=(node.input_values[0],), outputs=node.output_values)
    else:
        torch_module = OnnxInstanceNorm(momentum=momentum, epsilon=epsilon)
        onnx_mapping = onnx_mapping_from_node(node)

    return OperationConverterResult(torch_module=torch_module, onnx_mapping=onnx_mapping)



__all__ = [
    'OnnxMod',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node


class OnnxMod(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, fmod: int):
        super().__init__()
        self.fmod = fmod

        if self.fmod not in [0, 1]:
            raise ValueError(f'OnnxMod fom must be 0 or 1, but get {self.fmod}')

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return torch.fmod(x, y) if self.fmod else torch.remainder(x, y)


@add_converter(operation_type='Mod', version=10)
@add_converter(operation_type='Mod', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    fmod = node_attributes.get('fmod', 0)
    return OperationConverterResult(
        torch_module=OnnxMod(fmod=fmod),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



# pylint: disable=missing-docstring
__all__ = [
    'OnnxIsInf',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult


class OnnxIsInf(nn.Module, OnnxToTorchModule):
    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:
        return torch.isinf(input_tensor)


@add_converter(operation_type='IsInf', version=10)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    torch_module = OnnxIsInf()

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=OnnxMapping(
            inputs=(node.input_values[0],),
            outputs=node.output_values,
        ),
    )



__all__ = [
    'OnnxPadStatic',
    'OnnxPadDynamic',
]

from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import torch
import torch.nn.functional as F
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_const_value
from onnx2torch2.utils.common import onnx_mapping_from_node

_ONNX_TO_TORCH_MODE = {
    'constant': 'constant',
    'reflect': 'reflect',
    'edge': 'replicate',
}


def _onnx_to_torch_mode(mode: str) -> str:
    try:
        return _ONNX_TO_TORCH_MODE[mode]
    except KeyError as exc:
        raise NotImplementedError(f'{mode} mode is not implemented') from exc


def _torch_padding_to_mode_format(pads: List[int], mode: str) -> List[int]:
    if mode in ('replicate', 'reflect'):
        batch_channel_pads = pads[-4:]
        if set(batch_channel_pads) == {0}:
            return pads[:-4]

        raise RuntimeError(
            f'{mode} padding is implemented for padding the last 3 dimensions of 5D input tensor, '
            f'or the last 2 dimensions of 4D input tensor, or the last dimension of 3D input tensor.'
        )

    return pads


def _onnx_padding_to_torch(pads: List[int]) -> List[int]:
    # Convert padding from onnx format to torch format
    # onnx format: [x1_begin, x2_begin, ... , x1_end, x2_end, ...]
    # torch format [xn_begin, xn_end, ... , x2_begin, x2_end, x1_begin, x1_end]
    middle = len(pads) // 2
    onnx_pad_begin, onnx_pad_end = pads[:middle], pads[middle:]
    onnx_pad_begin, onnx_pad_end = onnx_pad_begin[::-1], onnx_pad_end[::-1]
    torch_pads = []
    for begin, end in zip(onnx_pad_begin, onnx_pad_end):
        torch_pads.extend([begin, end])

    return torch_pads


class OnnxPadStatic(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        pads: Union[Tuple[int, ...], List[int]],
        mode: str = 'constant',
        constant_value: float = 0.0,
    ):
        super().__init__()
        self.mode = mode
        self.pads = pads
        self.constant_value = constant_value

    @classmethod
    def create_from_onnx_params(  # pylint: disable=missing-function-docstring
        cls,
        onnx_pads: Union[Tuple[int, ...], List[int]],
        onnx_mode: str = 'constant',
        constant_value: float = 0.0,
    ) -> 'OnnxPadStatic':
        torch_mode = _onnx_to_torch_mode(onnx_mode)
        torch_padding = _onnx_padding_to_torch(onnx_pads)
        torch_padding = _torch_padding_to_mode_format(torch_padding, torch_mode)
        return cls(pads=torch_padding, mode=torch_mode, constant_value=constant_value)

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return F.pad(  # pylint: disable=not-callable
            input_tensor,
            mode=self.mode,
            pad=self.pads,
            value=self.constant_value,
        )


class OnnxPadDynamic(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, pads: List[int], mode: str = 'constant'):
        super().__init__()
        self.pads = pads
        self.mode = mode

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        pads: torch.Tensor,
        constant_value: Optional[float] = 0.0,
    ) -> torch.Tensor:
        torch_pads = _onnx_padding_to_torch(pads.tolist() if not self.pads else self.pads)
        torch_pads = _torch_padding_to_mode_format(torch_pads, self.mode)

        return F.pad(input_tensor, mode=self.mode, pad=torch_pads, value=constant_value)  # pylint: disable=not-callable


@add_converter(operation_type='Pad', version=11)
@add_converter(operation_type='Pad', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    mode = node.attributes.get('mode', 'constant')
    mode = _onnx_to_torch_mode(mode)

    pads_name = node.input_values[1]
    pads = []
    if pads_name in graph.initializers or pads_name in graph._node_output_values:  # pylint: disable=W0212
        # try and get padding arg value
        try:
            pads = get_const_value(pads_name, graph).tolist()
        except KeyError as exc:
            raise ValueError(
                f"Dynamic padding is not supported. Pad node `{node.name}`"
                f"has padding arg `{pads_name}` with no static value."
            ) from exc

    return OperationConverterResult(
        torch_module=OnnxPadDynamic(pads=pads, mode=mode),
        onnx_mapping=OnnxMapping(
            inputs=node.input_values,
            outputs=node.output_values,
        ),
    )


@add_converter(operation_type='Pad', version=2)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    torch_module = OnnxPadStatic.create_from_onnx_params(
        onnx_mode=node.attributes.get('mode', 'constant'),
        onnx_pads=node.attributes.get('pads'),
        constant_value=node.attributes.get('constant_value', 0.0),
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



from typing import Optional
from typing import Tuple
from typing import Union

from torch import nn

from onnx2torch2.node_converters.pad import OnnxPadStatic


def is_symmetric_onnx_padding(padding: Tuple[int, ...]) -> bool:  # pylint: disable=missing-function-docstring
    half_len = len(padding) // 2
    return padding[:half_len] == padding[half_len:]


def onnx_auto_pad_to_torch_padding(  # pylint: disable=missing-function-docstring
    auto_pad: str,
    onnx_padding: Tuple[int, ...],
) -> Tuple[Union[int, Tuple[int, ...]], Optional[nn.Module]]:
    if auto_pad == 'NOTSET':
        if onnx_padding is None:
            return 0, None

        if is_symmetric_onnx_padding(onnx_padding):
            half_len = len(onnx_padding) // 2
            return onnx_padding[:half_len], None

        return 0, OnnxPadStatic.create_from_onnx_params(onnx_pads=onnx_padding)

    if auto_pad == 'VALID':
        return 0, None

    if auto_pad in ('SAME_UPPER', 'SAME_LOWER'):
        raise NotImplementedError(f'"{auto_pad}" auto_pad is not implemented')

    raise ValueError(f'Got unexpected auto_pad value "{auto_pad}"')



# pylint: disable=missing-class-docstring
__all__ = [
    'OnnxReduceSumDynamicAxes',
    'OnnxReduceSumStaticAxes',
    'OnnxReduceStaticAxes',
]

from functools import partial
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OnnxMapping
from onnx2torch2.utils.common import OnnxToTorchModule
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_const_value
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


@torch.fx.wrap
def _get_element(x: Any, index: int = 0) -> Any:
    if isinstance(x, (tuple, list)):
        return x[index]

    return x


def _initialize_none_dim(
    dim: Optional[Union[int, Tuple[int, ...]]],
    input_dim: int,
) -> Union[List[int], Tuple[int, ...], int]:
    if dim is None:
        return list(range(input_dim))

    return dim


def _log_sum(
    input_tensor: torch.Tensor,
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False,
) -> torch.Tensor:
    dim_ = _initialize_none_dim(dim, input_tensor.dim())
    return torch.log(torch.sum(input_tensor, dim=dim_, keepdim=keepdim))


def _log_sum_exp(
    input_tensor: torch.Tensor,
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False,
) -> torch.Tensor:
    dim_ = _initialize_none_dim(dim, input_tensor.dim())
    return torch.logsumexp(input_tensor, dim=dim_, keepdim=keepdim)


def _sum_square(
    input_tensor: torch.Tensor,
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False,
) -> torch.Tensor:
    dim_ = _initialize_none_dim(dim, input_tensor.dim())
    return torch.sum(torch.square(input_tensor), dim=dim_, keepdim=keepdim)


_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'ReduceL1': partial(torch.norm, p=1),
    'ReduceL2': partial(torch.norm, p=2),
    'ReduceLogSum': _log_sum,
    'ReduceLogSumExp': _log_sum_exp,
    'ReduceMax': torch.max,
    'ReduceMean': torch.mean,
    'ReduceMin': torch.min,
    'ReduceProd': torch.prod,
    'ReduceSum': torch.sum,
    'ReduceSumSquare': _sum_square,
}


class OnnxReduceSumDynamicAxes(nn.Module, OnnxToTorchModuleWithCustomExport):
    def __init__(self, keepdims: int = 1, noop_with_empty_axes: int = 0):
        super().__init__()

        self._keepdims = keepdims
        self._noop_with_empty_axes = noop_with_empty_axes

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        del opset_version
        return {
            'noop_with_empty_axes_i': self._noop_with_empty_axes,
            'keepdims_i': self._keepdims,
        }

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        axes: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        def _forward() -> torch.Tensor:
            if axes is None or axes.nelement() == 0:
                if self._noop_with_empty_axes:
                    return input_tensor

                if not self._keepdims:
                    return torch.sum(input_tensor)

                fixed_axes = list(range(input_tensor.dim()))
            else:
                fixed_axes = torch.sort(axes).values.tolist()

            return torch.sum(input_tensor, dim=fixed_axes, keepdim=bool(self._keepdims))

        if torch.onnx.is_in_onnx_export():
            args = [input_tensor]
            if axes is not None:
                args.append(axes)

            onnx_attrs = self._onnx_attrs(opset_version=get_onnx_version())
            return DefaultExportToOnnx.export(_forward, 'ReduceSum', *args, onnx_attrs)

        return _forward()


class OnnxReduceSumStaticAxes(nn.Module, OnnxToTorchModule):
    def __init__(
        self,
        axes: List[int],
        keepdims: int = 1,
        noop_with_empty_axes: int = 0,
    ):
        super().__init__()
        if axes is not None:
            axes = sorted(axes)

        self._keepdims = bool(keepdims)
        self._noop_with_empty_axes = noop_with_empty_axes
        self._axes = axes

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self._axes is None or len(self._axes) == 0:
            if self._noop_with_empty_axes:
                return input_tensor

            if not self._keepdims:
                return self.math_op_function(input_tensor)

            self._axes = list(range(input_tensor.dim()))

        return torch.sum(input_tensor, dim=self._axes, keepdim=bool(self._keepdims))


class OnnxReduceStaticAxes(nn.Module, OnnxToTorchModule):
    def __init__(
        self,
        operation_type: str,
        axes: Optional[List[int]],
        keepdims: int = 1,
    ):
        super().__init__()
        self.operation_type = operation_type
        self.math_op_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

        if axes is not None:
            axes = sorted(axes)

        self.keepdims = keepdims == 1
        self.axes = axes

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self.axes is None or len(self.axes) == 0:
            if not self.keepdims:
                return self.math_op_function(input_tensor)

            self.axes = list(range(input_tensor.dim()))

        if self.operation_type not in ['ReduceMax', 'ReduceMin', 'ReduceProd']:
            return self.math_op_function(input_tensor, dim=self.axes, keepdim=self.keepdims)

        result = input_tensor
        for passed_dims, axis in enumerate(self.axes):
            result = self.math_op_function(
                result,
                dim=axis if self.keepdims else axis - passed_dims,
                keepdim=self.keepdims,
            )
            result = _get_element(result, 0)

        return result


@add_converter(operation_type='ReduceL1', version=1)
@add_converter(operation_type='ReduceL1', version=11)
@add_converter(operation_type='ReduceL1', version=13)
@add_converter(operation_type='ReduceL2', version=1)
@add_converter(operation_type='ReduceL2', version=11)
@add_converter(operation_type='ReduceL2', version=13)
@add_converter(operation_type='ReduceLogSum', version=1)
@add_converter(operation_type='ReduceLogSum', version=11)
@add_converter(operation_type='ReduceLogSum', version=13)
@add_converter(operation_type='ReduceLogSumExp', version=1)
@add_converter(operation_type='ReduceLogSumExp', version=11)
@add_converter(operation_type='ReduceLogSumExp', version=13)
@add_converter(operation_type='ReduceMax', version=1)
@add_converter(operation_type='ReduceMax', version=11)
@add_converter(operation_type='ReduceMax', version=12)
@add_converter(operation_type='ReduceMax', version=13)
@add_converter(operation_type='ReduceMean', version=1)
@add_converter(operation_type='ReduceMean', version=11)
@add_converter(operation_type='ReduceMean', version=13)
@add_converter(operation_type='ReduceMin', version=1)
@add_converter(operation_type='ReduceMin', version=11)
@add_converter(operation_type='ReduceMin', version=12)
@add_converter(operation_type='ReduceMin', version=13)
@add_converter(operation_type='ReduceProd', version=1)
@add_converter(operation_type='ReduceProd', version=11)
@add_converter(operation_type='ReduceProd', version=13)
@add_converter(operation_type='ReduceSum', version=1)
@add_converter(operation_type='ReduceSum', version=11)
@add_converter(operation_type='ReduceSumSquare', version=1)
@add_converter(operation_type='ReduceSumSquare', version=11)
@add_converter(operation_type='ReduceSumSquare', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    del graph
    node_attributes = node.attributes
    axes: Optional[List[int]] = node_attributes.get('axes', None)
    keepdims: int = node_attributes.get('keepdims', 1)

    return OperationConverterResult(
        torch_module=OnnxReduceStaticAxes(
            operation_type=node.operation_type,
            axes=axes,
            keepdims=keepdims,
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='ReduceSum', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    keepdims: int = node.attributes.get('keepdims', 1)
    noop_with_empty_axes: int = node.attributes.get('noop_with_empty_axes', 0)

    if len(node.input_values) == 2:
        try:
            axes = cast(torch.Tensor, get_const_value(node.input_values[1], graph))
            axes = axes.tolist()
            return OperationConverterResult(
                torch_module=OnnxReduceSumStaticAxes(
                    axes=axes,
                    keepdims=keepdims,
                    noop_with_empty_axes=noop_with_empty_axes,
                ),
                onnx_mapping=OnnxMapping(
                    inputs=(node.input_values[0],),
                    outputs=node.output_values,
                ),
            )
        except KeyError:
            pass

    return OperationConverterResult(
        torch_module=OnnxReduceSumDynamicAxes(keepdims=keepdims, noop_with_empty_axes=noop_with_empty_axes),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxRoiAlign',
]

from enum import Enum
from typing import Any
from typing import Dict

import torch
from torch import nn
from torchvision.ops import roi_align

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import get_onnx_version
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class CoordinateTransformationModeOnnxAttr(Enum):
    """
    Representation of new attribute in 16 opset: `coordinate_transformation_mode`.

    Allowed values are `half_pixel` and `output_half_pixel`.
    Use the value `half_pixel` to pixel shift the input coordinates by -0.5 (the recommended behavior).
    Use the value `output_half_pixel` to omit the pixel shift for the inputs
    (use this for a backward-compatible behavior).
    """

    HALF_PIXEL = 'half_pixel'
    OUTPUT_HALF_PIXEL = 'output_half_pixel'


CTMOnnxAttr = CoordinateTransformationModeOnnxAttr  # Type alias.


class OnnxRoiAlign(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        coordinate_transformation_mode: CTMOnnxAttr = CTMOnnxAttr.HALF_PIXEL,
        mode: str = 'avg',
        output_height: int = 1,
        output_width: int = 1,
        sampling_ratio: int = 0,
        spatial_scale: float = 1.0,
    ):
        super().__init__()

        self._coordinate_transformation_mode = coordinate_transformation_mode

        if mode != 'avg':
            raise NotImplementedError(f'"{mode}" roi align mode is not implemented.')
        self._mode = mode

        self._output_height = output_height
        self._output_width = output_width
        self._sampling_ratio = sampling_ratio
        self._spatial_scale = spatial_scale

    def _onnx_attrs(self, opset_version: int) -> Dict[str, Any]:
        onnx_attrs: Dict[str, Any] = {
            'mode_s': self._mode,
            'output_height_i': self._output_height,
            'output_width_i': self._output_width,
            'sampling_ratio_i': self._sampling_ratio,
            'spatial_scale_f': self._spatial_scale,
        }

        if opset_version < 16:
            if self._coordinate_transformation_mode != CTMOnnxAttr.OUTPUT_HALF_PIXEL:
                raise ValueError(
                    'RoiAlign from opset 10 does not support coordinate_transform_mode != "output_half_pixel"'
                    f', got {self._coordinate_transformation_mode.value}'
                )
            return onnx_attrs

        onnx_attrs['coordinate_transformation_mode_s'] = self._coordinate_transformation_mode.value
        return onnx_attrs

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        rois: torch.Tensor,
        batch_indices: torch.Tensor,
    ) -> torch.Tensor:
        def _forward():
            fixed_batch_indices = batch_indices.unsqueeze(1).to(rois.dtype)
            batched_rois = torch.cat([fixed_batch_indices, rois], dim=1)
            output_size = self._output_height, self._output_width
            sampling_ratio = self._sampling_ratio
            spatial_scale = self._spatial_scale

            return roi_align(
                input=input_tensor,
                boxes=batched_rois,
                output_size=output_size,
                spatial_scale=spatial_scale,
                sampling_ratio=sampling_ratio,
                aligned=self._coordinate_transformation_mode == CTMOnnxAttr.HALF_PIXEL,
            )

        if torch.onnx.is_in_onnx_export():
            onnx_attrs = self._onnx_attrs(get_onnx_version())
            return DefaultExportToOnnx.export(_forward, 'RoiAlign', input_tensor, rois, batch_indices, onnx_attrs)

        return _forward()


def converter_schema(  # pylint: disable=missing-function-docstring, unused-argument
    node: OnnxNode,
    graph: OnnxGraph,
    default_ctm: str,
) -> OperationConverterResult:
    node_attributes = node.attributes
    coordinate_transformation_mode = CTMOnnxAttr(node_attributes.get('coordinate_transformation_mode', default_ctm))
    mode = node_attributes.get('mode', 'avg')
    output_height = node_attributes.get('output_height', 1)
    output_width = node_attributes.get('output_width', 1)
    sampling_ratio = node_attributes.get('sampling_ratio', 0)
    spatial_scale = node_attributes.get('spatial_scale', 1.0)

    return OperationConverterResult(
        torch_module=OnnxRoiAlign(
            coordinate_transformation_mode=coordinate_transformation_mode,
            mode=mode,
            output_height=output_height,
            output_width=output_width,
            sampling_ratio=sampling_ratio,
            spatial_scale=spatial_scale,
        ),
        onnx_mapping=onnx_mapping_from_node(node),
    )


@add_converter(operation_type='RoiAlign', version=10)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return converter_schema(node=node, graph=graph, default_ctm='output_half_pixel')


@add_converter(operation_type='RoiAlign', version=16)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return converter_schema(node=node, graph=graph, default_ctm='half_pixel')



__all__ = [
    'OnnxReshape',
]

import torch
from torch import nn

from onnx2torch2.node_converters.registry import add_converter
from onnx2torch2.onnx_graph import OnnxGraph
from onnx2torch2.onnx_node import OnnxNode
from onnx2torch2.utils.common import OperationConverterResult
from onnx2torch2.utils.common import onnx_mapping_from_node
from onnx2torch2.utils.custom_export_to_onnx import DefaultExportToOnnx
from onnx2torch2.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxReshape(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    @staticmethod
    def _do_reshape(input_tensor: torch.Tensor, shape: torch.Tensor) -> torch.Tensor:
        if torch.any(shape == 0):
            shape = [input_tensor.shape[i] if dim_size == 0 else dim_size for i, dim_size in enumerate(shape)]

        return torch.reshape(input_tensor, torch.Size(shape))

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        shape: torch.Tensor,
    ) -> torch.Tensor:
        def _forward() -> torch.Tensor:
            return self._do_reshape(input_tensor, shape)

        if torch.onnx.is_in_onnx_export():
            return DefaultExportToOnnx.export(_forward, 'Reshape', input_tensor, shape, {})

        return _forward()


@add_converter(operation_type='Reshape', version=5)
@add_converter(operation_type='Reshape', version=13)
@add_converter(operation_type='Reshape', version=14)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    if node.attributes.get('allowzero', 0) == 1:
        raise NotImplementedError('"allowzero=1" is not implemented')

    return OperationConverterResult(
        torch_module=OnnxReshape(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )
